/*
 * BIG BROTHER DAEMON
 * Sean MacGuire - The MacLawran Group Inc.
 *
 * This daemon just listens on a port, accepts single lines of input
 * and handles them.  It doesn't send any acknowledgements back...
 *
 * Two types of actions are possible, depending on the content we get:
 *
 *	1.	Input for the display monitor
 *	2.	Request to send a page to the admin
 */

#include "bb.h"		/* THE BIG BROTHER INCLUDE FILE */
#include <errno.h>

char bblogdir[MAXLINE];		/* DIRECTORY TO PUT LOG FILES INTO */
char pager[MAXLINE];		/* BB-PAGER PROGRAM LOCATION */

main()
{
 	int status, sockfd, newsockfd, child, port, len;
	void cleanup();
        struct sockaddr_in serv_addr, cli_addr;
	static char line[MAXLINE];
	char *bbhome;				/* WHERE ALL BB LIVES */

	bbhome=(char *)getenv("BBHOME");
	if (bbhome) {
		sprintf(bblogdir, "%s/www/logs", bbhome);
		sprintf(pager, "%s/bin/bb-page.sh", bbhome);
	}

	port = PORT;

        debug("MAKING SOCKET CONNECTION...\n");
        if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
                fprintf(stderr, "Can't open stream socket");
		exit(1);
        }
        debug("SOCKET CONNECTED OK...\n");

        bzero((char *) &serv_addr, sizeof(serv_addr));
        serv_addr.sin_family = AF_INET;
        serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        serv_addr.sin_port = htons(port);

        debug("BINDING TO SOCKET...\n");
        if (bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0) {
                fprintf(stderr,"Can't open stream socket");
		exit(2);
        }
        debug("SOCKET BOUND OK...\n");

        /*
	 * NOW THE PARENT EXITS AND THE CHILD DETACHES FROM THE CONTROLLING TTY
	 * SET UP TO IGNORE SIGNALS SENT BY EXITING CHILDREN
         */
        if ((child = fork()) != 0) exit(0);             /* PARENT EXITS */
	setsid();					/* RELEASE TTY */

#ifndef ZOMBIE
        signal(SIGCHLD, SIG_IGN);
#endif /* NO ZOMBIE */

	/*
	 * NOW SET UP THE LISTENING LOOP
         */
        listen(sockfd, 5);

        for ( ;; ) {
                len = sizeof(cli_addr);
                newsockfd = accept(sockfd,(struct sockaddr*)&cli_addr,&len);
		/*
		 * THANKS TO Marc L. Allen (allen@allen@chesapeake.rps.slb.com)
		 */
		if (newsockfd < 0 ) {
			if (errno == EINTR) continue;
                	else die("server: accept error");
		}

		/*
		 * NOW FORK OFF AND LET THE CHILD HANDLE THE REQUEST
		 */
 		if ((child = fork()) < 0) die("server: fork error");

		if (child != 0) {			/* THE PARENT */
			close(newsockfd);
#ifdef ZOMBIE
			debug("WAIT FOR CHILD EXIT...\n");
			wait(&status);
			debug("CLEANED UP CHILD: %d\n", status);
#endif /* ZOMBIE */
		}
		else {					/* THE CHILD */
			close(sockfd); 
			/*
			 * NOW GET THE LINE FROM THE CLIENT
			 * AND HANDLE THE REQUEST...
			 * IGNORE BAD DATA...
			 */
			len = recv(newsockfd, line, MAXLINE, 0);     
			if (len > 0) do_bb(line);
			close(newsockfd);
                	exit(0);
		}
	}
}

/*
 * HANDLE INCOMING BIG BROTHER REQUESTS
 * PARSE THE LINE WE'VE RECEIVED AND HANDLE IT
 */
do_bb(req)
char *req;
{
	debug("REC'D: %s\n", req);
	/*
	 * IF THE REQUEST STARTS WITH "page" THEN CALL THE PAGING SCRIPT
	 * IF THE REQUEST STARTS WITH "status" THEN LOG THE MESSAGE
	 * OTHERWISE IGNORE IT (FOR SAFETY SAKE)
	 */
	if (strncmp(req, "page ", 5) == 0) {
		req += 5;			/* JUMP OVER page */
		debug("CALL PAGER: %s\n", req);
		if (pager) {
			execlp(pager, pager, req, (char *)NULL);
			die("SERVER PAGE GAGGED\n");
		}
		else die("ENVIONMENT VARIABLE BBHOME UNDEFINED!\n");
	}
	else if (strncmp(req, "status ", 7) == 0) {
		req += 7;			/* JUMP OVER status */
		debug("LOG STATUS MSG: %s\n", req);
		bblog(req);			/* LOG THE REQUEST */
	}
	else debug("IGNORE MESSAGE: %s\n", req);
}

/*
 * LOG A BIG BROTHER STATUS MESSAGE
 * THE MESSAGE GOES INTO THE DIRECTORY SPECIFIED BY BBLOG
 * MESSAGE FORMAT IS:
 * 	machine.area status message
 *
 * THE MESSAGE IS PLACED IN A FILE NAMED machine.area
 */

int bblog(msg)
char *msg;
{
        FILE *bbfile;
        int len;
        char fname[MAXLINE];				/* FILENAME */
        char *filename, *name;

        if (bblogdir) {
		filename=(char *)malloc(MAXLINE);      /* MACHINE NAME */
                sscanf(msg, "%s", filename);
                /* 
		 * THANKS TO Jac Kersing <J.Kersing@the-box.com>
		 * SMM: IGNORE ATTEMPTS TO WRITE TO PARENT DIRECTORIES
		 * AND CHECK FOR MALLOC OVERFLOWS
                 */
                name = (char *)strrchr(filename,'/');
                if (name != (char *)NULL) {		/* FOUND A FUNNY */
			debug("ILLEGAL FILE NAME - IGNORED\n");
			return;				/* GO AWAY */
		}

                len = strlen(filename);                 /* FOR LATER */
                if (strlen(bblogdir) + strlen(filename) + 1 > MAXLINE)
                {
                        /* too large for allocated memory, skip it */
			debug("FILE NAME TOO BIG - IGNORED\n");
                        return;
                }

        	sprintf(fname, "%s/%s", bblogdir, filename);
        	debug("NAME IS %s....\n", fname);
        	debug("OPENING BBFILE....\n");
	
        	bbfile = fopen(fname, "w+");            /* APPEND IT */
        	msg += len + 1;                         /* SKIP FIRST ARGS */
        	debug("MSG IS %s....\n", msg);
        	fprintf(bbfile, "%s\n", msg);           /* PUKE */
        	fclose(bbfile);                         /* CLOSE */
        	free(filename);                         /* FREE MEM */
	}
	else die("ENV VAR BBHOME UNDEFINED\n");
}

/*
 * PUT AN ERROR INTO THE LOGFILE AND stderr AND DIE
 */
int die(msg)
char *msg;
{
        static char message[MAXLINE];
        sprintf(message, "bbd: %s\n", msg);

        syslog(LOG_DAEMON | LOG_CRIT , message);
        fprintf(stderr, message);
        exit(1);
}
